import chalk from 'chalk';
import fs, { existsSync, statSync, readdirSync, readFileSync } from 'node:fs';
import { sep, join } from 'node:path';
import esbuild from 'esbuild';
import { createHash } from 'node:crypto';

const isObject = (v) => {
  return Object.prototype.toString.call(v) === "[object Object]";
};
const getFileNameOrFileDir = (path) => {
  if (existsSync(path) && statSync(path)?.isFile()) {
    const arrPath = path.split(sep);
    return {
      filename: arrPath.slice(-1)[0],
      fileDir: arrPath.slice(0, arrPath.length - 1).join(sep)
    };
  } else {
    return { filename: "index", fileDir: path };
  }
};
const logError = (...args) => {
  console.log(chalk.red(...args));
};
const logWarn = (...args) => {
  console.log(chalk.yellow(...args));
};
const pathToStr = (path) => {
  if (typeof path === "string")
    return path;
  if (Array.isArray(path))
    return path.join();
  throw new Error(`Error, path \u7C7B\u578B\u9519\u8BEF\uFF01`);
};
const cleanFields = (obj) => {
  if (Array.isArray(obj)) {
    return obj.reduce((acc, cur) => {
      if (cur !== void 0) {
        if (isObject(cur) || Array.isArray(cur)) {
          acc.push(cleanFields(cur));
        } else {
          acc.push(cur);
        }
      }
      return acc;
    }, []);
  }
  if (isObject(obj)) {
    return Object.entries(obj).reduce((acc, [k, v]) => {
      if (v !== void 0) {
        if (isObject(v) || Array.isArray(v)) {
          acc = { ...acc, [k]: cleanFields(v) };
        } else {
          acc = { ...acc, [k]: v };
        }
      }
      return acc;
    }, {});
  }
  return obj;
};
const detectUnique = (data) => {
  const map = /* @__PURE__ */ new Map();
  function recursive(data2) {
    for (let i = 0, len = data2.length; i < len; i += 1) {
      const node = data2[i];
      const key = node.key;
      map.set(key, map.has(key) ? (map.get(key) ?? 0) + 1 : 1);
      if (node.children) {
        node.children = recursive(node.children);
      }
    }
    return data2;
  }
  recursive(data);
  return map;
};
function exit(x) {
  if (isObject(x) && x.msg) {
    logError(x.msg?.stack ?? x.msg);
    process.exit(x.code ?? -1);
  } else {
    logError(x?.stack ?? x);
    process.exit(-1);
  }
}

const md5 = (data) => createHash("md5").update(data).digest("hex");
const path2key = (path) => {
  let key = "";
  if (!path)
    return key;
  key = path.replace(/^\//, "").replace(/\//g, ":");
  if (/,/.test(key)) {
    const tmpKeys = key.split(",");
    const first = tmpKeys[0];
    return [first, md5(path)].join(":").toUpperCase();
  }
  return key.toUpperCase().replace(/-/g, "_");
};
const getSpecifiedRouteIndex = (routes, route) => {
  for (let i = 0; i < routes.length; i += 1) {
    const item = routes[i];
    if (item.path === route.path && item.hideInMenu) {
      return i;
    }
  }
  return -1;
};
const validKey = (key) => {
  if (key.length > 128) {
    throw new Error(`\u5F02\u5E38\uFF1Akey\u3010${key}\u3011\u957F\u5EA6\u8D85\u8FC7\u9650\u5236(128)`);
  }
  return key;
};
const validPath = (path) => {
  if (path.length > 1e4) {
    throw new Error(`\u5F02\u5E38\uFF1Apath\u3010${path}\u3011\u957F\u5EA6\u8D85\u8FC7\u9650\u5236(10000)`);
  }
  return path;
};

const parseAuth = ({ apiPrefix, auth }) => {
  let path = auth.path ? pathToStr(auth.path) : void 0;
  const key = validKey(auth.authKey || auth.key || path2key(path));
  if (!auth.authKey && !auth.key) {
    logWarn(
      `[WARN]: \u6743\u9650\u914D\u7F6E\u9879\u3010${JSON.stringify(auth)}\u3011\u6CA1\u6709\u624B\u52A8\u4F20\u3010key\u3011\uFF0C\u8FD9\u662F\u4E0D\u88AB\u671F\u671B\u7684\uFF01\uFF01\uFF01`
    );
  }
  apiPrefix = auth.apiPrefix ?? apiPrefix;
  if (path && apiPrefix) {
    if (!apiPrefix.startsWith("/")) {
      apiPrefix = "/" + apiPrefix;
    }
    path = path.split(",").map((c) => {
      const p = c.startsWith(apiPrefix) ? c : [apiPrefix, c].join("/");
      return p.replace(/\/+/g, "/").replace(/\/$/, "");
    }).join(",");
  }
  if (path)
    validPath(path);
  const type = auth.type || (path ? "api" : "component");
  return { ...auth, type, key, path };
};
const trans = ({ apiPrefix, data }) => {
  if (!data)
    return [];
  const parse = (data2) => {
    return data2.reduce((prev, curr) => {
      const { children } = curr;
      const authData = parseAuth({ apiPrefix, auth: curr });
      const authTreeSources = cleanFields({
        ...authData ?? {},
        children: children?.length ? parse(children) : void 0
      });
      if (Object.keys(authTreeSources).length) {
        prev.push(authTreeSources);
      }
      return prev;
    }, []);
  };
  return parse(data);
};

const scanDir = (dirPath, excludes) => {
  if (!existsSync(dirPath))
    return;
  const files = readdirSync(dirPath);
  if (!files?.length)
    return;
  return files.reduce((acc, cur) => {
    const filePath = join(dirPath, cur);
    if (excludes?.length) {
      for (let i = 0, len = excludes.length; i < len; i++) {
        const exclude = excludes[i];
        if (!new RegExp(exclude).test(cur)) {
          acc.push(filePath);
        }
      }
    } else {
      acc.push(filePath);
    }
    return acc;
  }, []);
};
const readFileContent = (filePath) => {
  if (!existsSync(filePath))
    return;
  const stat = statSync(filePath);
  if (!stat)
    return;
  if (stat.isFile()) {
    const text = readFileSync(filePath, "utf8");
    const json = JSON.parse(text);
    return Array.isArray(json) ? json : [json];
  }
};

const ignoreAuth = (route, cfg) => {
  const { path, redirect, children, disableAuth } = route;
  const isFilter = cfg.routeFilterRules?.some((c) => path.includes(c));
  return isFilter || !path || redirect || // children?.length ||
  disableAuth;
};
const combinePaths = (data, key, path) => {
  if (!data?.length)
    return { path, json: void 0 };
  const idx = data.findIndex((c) => c.key === key && c.type === "route");
  if (idx >= 0) {
    return {
      path: [path, data[idx].path].join(","),
      json: [...data.slice(0, idx), ...data.slice(idx + 1)]
    };
  }
  return { path, json: data };
};
function transRoutes(cfg) {
  const filePath = cfg.entry + ".ts";
  if (!existsSync(filePath))
    return [];
  const text = readFileSync(filePath, "utf8");
  const { code } = esbuild.transformSync(text, { loader: "ts", minify: true });
  const jsonText = /\[\{.*\}\];/.exec(code)?.[0];
  const json = new Function("return " + jsonText)();
  const parse = (routes, { pName, pPath } = {}) => {
    return (routes || []).reduce((prev, curr) => {
      if (ignoreAuth(curr, cfg))
        return prev;
      validPath(curr.path);
      let { element } = curr;
      const { path, children, name } = curr;
      const full = {
        name: pName ? [pName, name].join(".") : name,
        path: path.startsWith("/") ? path : "/" + [...pPath?.split("/") || [], ...path?.split("/") || []].filter(Boolean).join("/")
      };
      if (!element && children) {
        const specialRouteIndex = getSpecifiedRouteIndex(children, curr);
        const specialRoute = children[specialRouteIndex];
        if (specialRoute) {
          element = specialRoute.element;
          children.splice(specialRouteIndex, 1);
        }
      }
      const key = validKey(curr.authKey || path2key(full.path));
      let childAuth;
      if (children?.length) {
        childAuth = parse(children, {
          pName: full.name,
          pPath: full.path
        });
      } else {
        const currPath = join(cfg.cwd, "src/pages/", element);
        const { fileDir } = getFileNameOrFileDir(currPath);
        const content = readFileContent(join(fileDir, "auth.json"));
        const apiPrefix = cfg.apiPrefix;
        if (content) {
          childAuth = trans({ data: content, apiPrefix });
        } else {
          const collection = scanDir(join(fileDir, "auth"))?.map((filePath2) => {
            const fc = readFileContent(filePath2);
            return trans({ data: fc, apiPrefix });
          });
          childAuth = collection?.flat(1);
        }
      }
      const { path: p, json: json2 } = combinePaths(childAuth, key, full.path);
      const authTree = {
        name,
        key,
        type: "route",
        path: p,
        hidden: curr.hideInMenu,
        icon: curr.icon,
        children: json2
      };
      prev.push(authTree);
      return prev;
    }, []);
  };
  return parse(json.find((c) => c.path === "/")?.children || []);
}

function bootstrap(opts) {
  try {
    const options = {
      /** 默认参数 */
      apiPrefix: "",
      // 接口前缀
      routeFilterRules: ["404", "403", "500", "*"],
      /** 业务传入 */
      ...opts,
      entry: join(opts.cwd, "config/base")
    };
    const data = cleanFields([...transRoutes(options)]);
    const map = detectUnique(data);
    [...map.entries()].forEach(([key, count]) => {
      if (count > 1)
        logError(
          `\u3010ERROR\u3011\u6743\u9650 key\u3010${key}\u3011\u91CD\u590D\u3010${count}\u3011\u6B21\uFF0C\u8BF7\u4FDD\u6301\u6743\u9650 key \u7684\u552F\u4E00\u6027\uFF01\uFF01\uFF01`
        );
    });
    return data;
  } catch (e) {
    return exit(e);
  }
}

function extractAuth(options) {
  return {
    name: "vite-plugin-extract-auth",
    configureServer: (service) => {
      const root = service.config.root;
      const cacheDir = service.config.cacheDir;
      try {
        const source = bootstrap({ cwd: root, ...options });
        const authorizationFile = `${cacheDir}/authorization.ts`;
        fs.writeFile(
          authorizationFile,
          `export const APP_KEY = "${options?.appKey}";

export default ${JSON.stringify(source, null, 2)};`,
          "utf-8",
          () => {
          }
        );
        console.log(chalk.green(`Write authorize config file success`));
      } catch (e) {
        console.log(
          chalk.red(`Write authorize config file failed, error message: ${e}`)
        );
      }
    },
    closeBundle: () => {
    }
  };
}

export { extractAuth as default };
